"""Functional programming utilities and composable operations."""

from typing import TypeVar, Callable, Any, Iterable, Optional, List, Dict, Union
from functools import reduce, wraps, partial
from .validation import validate_type
from .timing import timing_block
from .logging import log_dict, log_stage_execution, StageLogger
from loguru import logger
import inspect
from typing import get_type_hints
from contextlib import contextmanager, nullcontext

T = TypeVar('T')
U = TypeVar('U')
V = TypeVar('V')

def with_logging(level: str = "DEBUG") -> Callable:
    """Set logging level for a specific function.
    
    Args:
        level: Logging level for this function
        
    Returns:
        Decorated function with specific logging level
        
    Example:
        >>> @with_logging(level="DEBUG")
        ... def process_data(data):
        ...     return data.upper()
    """
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapped(*args, **kwargs) -> Any:
            # Store original level
            original_level = logger.level
            try:
                # Set function-specific level
                logger.level = level
                result = func(*args, **kwargs)
                return result
            finally:
                # Restore original level
                logger.level = original_level
        return wrapped
    return decorator

def log_args(level: str = "DEBUG") -> Callable:
    """Log function arguments and return value.
    
    Args:
        level: Logging level
        
    Returns:
        Decorated function that logs its arguments and result
        
    Example:
        >>> @log_args(level="DEBUG")
        ... def add(x, y):
        ...     return x + y
    """
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapped(*args, **kwargs) -> Any:
            args_dict = {
                'args': args,
                'kwargs': kwargs,
            }
            log_dict(args_dict, level=level, title=f"{func.__name__} arguments")
            result = func(*args, **kwargs)
            log_stage_execution(func.__name__, args_dict, result, level=level)
            return result
        return wrapped
    return decorator

def compose(*functions: Callable) -> Callable:
    """Compose multiple functions from right to left.
    
    Args:
        *functions: Functions to compose
        
    Returns:
        Composed function
        
    Example:
        >>> f = compose(str.upper, str.strip)
        >>> f("  hello  ")
        "HELLO"
    """
    if not functions:
        return lambda x: x
    
    def _compose(f: Callable, g: Callable) -> Callable:
        return lambda x: f(g(x))
    return reduce(_compose, functions)

def pipe(*functions: Callable) -> Callable:
    """Compose multiple functions from left to right.
    
    Args:
        *functions: Functions to compose
        
    Returns:
        Composed function
        
    Example:
        >>> f = pipe(str.strip, str.upper)
        >>> f("  hello  ")
        "HELLO"
    """
    if not functions:
        return lambda x: x
    return compose(*reversed(functions))

def curry(func: Callable) -> Callable:
    """Curry a function to allow partial application.
    
    Args:
        func: Function to curry
        
    Returns:
        Curried function
        
    Example:
        >>> @curry
        ... def add(x, y):
        ...     return x + y
        >>> add(1)(2)
        3
    """
    sig = inspect.signature(func)
    param_names = list(sig.parameters)
    
    @wraps(func)
    def curried(*args, **kwargs):
        # Try to bind arguments to see if we have all required args
        try:
            bound = sig.bind(*args, **kwargs)
            bound.apply_defaults()
            return func(*bound.args, **bound.kwargs)
        except TypeError:
            # If we can't bind all arguments, create a partial application
            def inner(*more_args, **more_kwargs):
                # Convert positional args to keyword args based on parameter order
                new_kwargs = kwargs.copy()
                args_as_kwargs = {
                    param_names[i]: arg 
                    for i, arg in enumerate(args)
                    if param_names[i] not in new_kwargs
                }
                new_kwargs.update(args_as_kwargs)
                new_kwargs.update(more_kwargs)
                
                # Convert remaining positional args
                remaining_args = []
                used_params = set(new_kwargs)
                for arg in more_args:
                    for param in param_names:
                        if param not in used_params:
                            new_kwargs[param] = arg
                            used_params.add(param)
                            break
                    else:
                        remaining_args.append(arg)
                
                return curried(*remaining_args, **new_kwargs)
            return inner
    return curried

def maybe(func: Callable[[T], U]) -> Callable[[Optional[T]], Optional[U]]:
    """Convert a function to handle None values.
    
    Args:
        func: Function to wrap
        
    Returns:
        Wrapped function that handles None
        
    Example:
        >>> @maybe
        ... def double(x: int) -> int:
        ...     return x * 2
        >>> double(None)
        None
        >>> double(2)
        4
    """
    @wraps(func)
    def wrapped(value: Optional[T]) -> Optional[U]:
        if value is None:
            logger.debug(f"{func.__name__} received None, returning None")
            return None
            
        # Get expected input type from type hints
        hints = get_type_hints(func)
        if hints and next(iter(hints.values())) != Any:
            expected_type = next(iter(hints.values()))
            if not validate_type(value, expected_type):
                raise TypeError(f"Expected {expected_type}, got {type(value)}")
                
        result = func(value)
        logger.debug(f"{func.__name__}({value}) -> {result}")
        return result
    return wrapped

def memoize(func: Callable) -> Callable:
    """Memoize a function's results.
    
    Args:
        func: Function to memoize
        
    Returns:
        Memoized function
        
    Example:
        >>> @memoize
        ... def fibonacci(n):
        ...     if n < 2: return n
        ...     return fibonacci(n-1) + fibonacci(n-2)
    """
    cache: Dict = {}
    sig = inspect.signature(func)
    
    @wraps(func)
    def wrapped(*args, **kwargs) -> Any:
        # Bind arguments to parameters to normalize them
        bound = sig.bind(*args, **kwargs)
        bound.apply_defaults()
        
        # Convert args to hashable form
        hashable_args = tuple(
            tuple(arg) if isinstance(arg, list) else arg
            for arg in bound.args
        )
        hashable_kwargs = tuple(
            (k, tuple(v) if isinstance(v, list) else v)
            for k, v in sorted(bound.kwargs.items())
        )
        key = (hashable_args, hashable_kwargs)
        
        if key in cache:
            logger.debug(f"{func.__name__} cache hit for {key}")
            return cache[key]
            
        logger.debug(f"{func.__name__} cache miss for {key}")
        cache[key] = func(*args, **kwargs)
        return cache[key]
    return wrapped

def trace(
    func: Optional[Callable] = None,
    *,
    level: Optional[str] = None,  # None means logging is off by default
    timing: bool = True,
    args: bool = True,
    types: bool = True,
    nested: bool = False
) -> Callable:
    """Trace function execution with configurable logging and timing.
    
    A powerful decorator for monitoring function execution. By default, it only tracks
    timing. Enable full tracing by setting a log level.
    
    Args:
        func: Function to trace (for use as @trace)
        level: Logging level (DEBUG, INFO, etc). None means no logging
        timing: Whether to log execution time
        args: Whether to log function arguments and results
        types: Whether to log type information
        nested: Whether to trace nested function calls
        
    Returns:
        Traced function
        
    Examples:
        Simple timing only:
        >>> @trace
        ... def process(data):
        ...     return data.upper()
        
        Full tracing with DEBUG level:
        >>> @trace(level="DEBUG")
        ... def process(data):
        ...     return data.upper()
        
        Custom configuration:
        >>> @trace(level="INFO", timing=True, args=False)
        ... def process(data):
        ...     return data.upper()
        
        Class method tracing:
        >>> class DataProcessor:
        ...     @trace(level="DEBUG")
        ...     def process(self, data):
        ...         return data.upper()
    """
    def decorator(func: Callable) -> Callable:
        # Get function signature and type hints once
        sig = inspect.signature(func)
        hints = get_type_hints(func) if types else {}
        
        def make_serializable(obj: Any) -> Any:
            """Convert non-serializable objects to string representation."""
            if isinstance(obj, (int, float, str, bool, type(None))):
                return obj
            if isinstance(obj, (list, tuple)):
                return [make_serializable(x) for x in obj]
            if isinstance(obj, dict):
                return {str(k): make_serializable(v) for k, v in obj.items()}
            return str(obj)
        
        @wraps(func)
        def wrapped(*args, **kwargs) -> Any:
            func_name = func.__name__
            
            # If no logging requested, just handle timing
            if level is None:
                if timing:
                    with timing_block(func_name):
                        return func(*args, **kwargs)
                return func(*args, **kwargs)
            
            # Full tracing
            try:
                # Prepare argument info
                if args:
                    # Bind args to parameter names
                    bound = sig.bind(*args, **kwargs)
                    bound.apply_defaults()
                    
                    # Create a readable argument representation
                    args_info = {
                        name: make_serializable(value)
                        for name, value in bound.arguments.items()
                    }
                    
                    # Log arguments
                    log_dict(
                        args_info, 
                        level=level,
                        title=f"{func_name} called with"
                    )
                
                # Log type information
                if types and hints:
                    type_info = {
                        "params": {
                            name: str(hint) 
                            for name, hint in hints.items() 
                            if name != "return"
                        },
                        "return": str(hints.get("return", "Any"))
                    }
                    log_dict(
                        type_info,
                        level=level,
                        title=f"{func_name} type information"
                    )
                
                # Execute with timing
                logger.log(level, f"Executing {func_name}")
                with timing_block(func_name, level=level) if timing else nullcontext():
                    result = func(*args, **kwargs)
                
                # Log result
                if args:
                    log_stage_execution(
                        func_name,
                        args_info,
                        make_serializable(result),
                        level=level
                    )
                
                return result
                
            except Exception as e:
                logger.log(
                    level,
                    f"Error in {func_name}: {e.__class__.__name__}({str(e)})"
                )
                raise
                
        return wrapped
    
    # Handle both @trace and @trace() syntax
    if func is None:
        return decorator
    return decorator(func)

class Pipeline:
    """A pipeline for chaining operations with validation.
    
    Example:
        >>> pipeline = Pipeline()\\
        ...     .map(str.strip)\\
        ...     .filter(bool)\\
        ...     .map(str.upper)
        >>> list(pipeline.process(["  ", "hello", " world "]))
        ["HELLO", "WORLD"]
    """
    
    def __init__(self, name: str = "pipeline", level: str = "DEBUG"):
        self.operations: List[Callable] = []
        self.name = name
        self.level = level
    
    def map(self, func: Callable[[T], U]) -> 'Pipeline':
        """Add a map operation to the pipeline."""
        logger.log(self.level, f"{self.name}: Adding map operation with {func.__name__}")
        self.operations.append(lambda data: map(func, data))
        return self
    
    def filter(self, predicate: Callable[[T], bool]) -> 'Pipeline':
        """Add a filter operation to the pipeline."""
        logger.log(self.level, f"{self.name}: Adding filter operation with {predicate.__name__}")
        self.operations.append(lambda data: filter(predicate, data))
        return self
    
    def reduce(self, func: Callable[[T, T], T], initial: Optional[T] = None) -> 'Pipeline':
        """Add a reduce operation to the pipeline."""
        logger.log(self.level, f"{self.name}: Adding reduce operation with {func.__name__}")
        if initial is not None:
            self.operations.append(lambda data: reduce(func, data, initial))
        else:
            self.operations.append(lambda data: reduce(func, data))
        return self
    
    def validate(self, type_: type) -> 'Pipeline':
        """Add type validation to the pipeline."""
        logger.log(self.level, f"{self.name}: Adding validation for type {type_.__name__}")
        self.operations.append(
            lambda data: filter(lambda x: validate_type(x, type_), data)
        )
        return self
    
    def process(self, data: Iterable[T]) -> Iterable[Any]:
        """Process data through the pipeline."""
        logger.log(self.level, f"{self.name}: Starting pipeline processing")
        result = data
        for i, operation in enumerate(self.operations, 1):
            with timing_block(f"{self.name} - step {i}", level=self.level):
                result = operation(result)
        logger.log(self.level, f"{self.name}: Pipeline processing complete")
        return result 