"""I/O utilities for common file operations."""

import os
import json
from pathlib import Path
from typing import Union, Dict, Any, Optional, Generator
from PIL import Image
import tempfile
from contextlib import contextmanager
from loguru import logger

PathLike = Union[str, Path]

def ensure_path(path: PathLike) -> Path:
    """Convert a path-like object to a Path and ensure it exists.
    
    Args:
        path: Path-like object
        
    Returns:
        Path object
        
    Raises:
        FileNotFoundError: If the path doesn't exist
    """
    path = Path(path)
    if not path.exists():
        raise FileNotFoundError(f"Path does not exist: {path}")
    return path

def load_json(path: PathLike) -> Dict[str, Any]:
    """Load JSON from a file.
    
    Args:
        path: Path to JSON file
        
    Returns:
        Loaded JSON data
        
    Raises:
        FileNotFoundError: If the file doesn't exist
        json.JSONDecodeError: If the file is not valid JSON
    """
    path = ensure_path(path)
    logger.debug(f"Loading JSON from {path}")
    with open(path, 'r', encoding='utf-8') as f:
        return json.load(f)

def save_json(data: Dict[str, Any], path: PathLike, indent: int = 2) -> None:
    """Save data as JSON to a file.
    
    Args:
        data: Data to save
        path: Path to save to
        indent: Number of spaces for indentation
    """
    path = Path(path)
    logger.debug(f"Saving JSON to {path}")
    with open(path, 'w', encoding='utf-8') as f:
        json.dump(data, f, indent=indent)

def load_image(path: PathLike, convert_to_rgb: bool = True) -> Image.Image:
    """Load an image from a file.
    
    Args:
        path: Path to image file
        convert_to_rgb: Whether to convert to RGB mode
        
    Returns:
        Loaded image
        
    Raises:
        FileNotFoundError: If the file doesn't exist
        PIL.UnidentifiedImageError: If the file is not a valid image
    """
    path = ensure_path(path)
    logger.debug(f"Loading image from {path}")
    image = Image.open(path)
    if convert_to_rgb and image.mode != 'RGB':
        image = image.convert('RGB')
    return image

def save_image(image: Image.Image, path: PathLike, format: Optional[str] = None) -> None:
    """Save an image to a file.
    
    Args:
        image: Image to save
        path: Path to save to
        format: Image format (inferred from path if None)
    """
    path = Path(path)
    logger.debug(f"Saving image to {path}")
    image.save(path, format=format)

@contextmanager
def temp_path(suffix: Optional[str] = None) -> Generator[Path, None, None]:
    """Create a temporary file path that is automatically cleaned up.
    
    Args:
        suffix: Optional file extension
        
    Yields:
        Path to temporary file
    """
    fd, path = tempfile.mkstemp(suffix=suffix)
    try:
        os.close(fd)  # Close file descriptor
        path = Path(path)
        logger.debug(f"Created temporary file: {path}")
        yield path
    finally:
        try:
            if path.exists():
                path.unlink()
                logger.debug(f"Cleaned up temporary file: {path}")
        except Exception as e:
            logger.warning(f"Failed to clean up temporary file {path}: {e}")

def get_file_info(path: PathLike) -> Dict[str, Any]:
    """Get information about a file.
    
    Args:
        path: Path to file
        
    Returns:
        Dictionary with file information
    """
    path = ensure_path(path)
    stats = path.stat()
    return {
        'name': path.name,
        'extension': path.suffix,
        'size': stats.st_size,
        'created': stats.st_ctime,
        'modified': stats.st_mtime,
        'is_file': path.is_file(),
        'is_dir': path.is_dir()
    } 