"""Logging utilities for consistent logging across the package."""

import sys
from typing import Optional, Dict, Any
from loguru import logger
import json

def configure_logging(
    level: str = "INFO",
    format: str = "<green>{time:YYYY-MM-DD HH:mm:ss}</green> | "
                 "<level>{level: <8}</level> | "
                 "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - "
                 "<level>{message}</level>",
    **kwargs: Any
) -> None:
    """Configure logging for the package.
    
    Args:
        level: Logging level (DEBUG, INFO, WARNING, ERROR, CRITICAL)
        format: Log message format
        **kwargs: Additional configuration for loguru
    """
    # Remove default handler
    logger.remove()
    
    # Add new handler with configuration
    logger.add(
        sys.stderr,
        format=format,
        level=level,
        **kwargs
    )

def log_dict(
    data: Dict[str, Any],
    level: str = "DEBUG",
    title: Optional[str] = None
) -> None:
    """Log a dictionary in a pretty format.
    
    Args:
        data: Dictionary to log
        level: Logging level
        title: Optional title for the log message
    """
    message = json.dumps(data, indent=2)
    if title:
        message = f"{title}:\n{message}"
    logger.log(level, message)

def log_stage_execution(
    stage_name: str,
    input_data: Any,
    output_data: Any,
    level: str = "DEBUG"
) -> None:
    """Log stage execution details.
    
    Args:
        stage_name: Name of the stage
        input_data: Input data to the stage
        output_data: Output data from the stage
        level: Logging level
    """
    logger.log(level, f"Executing stage: {stage_name}")
    logger.log(level, f"Input type: {type(input_data)}")
    logger.log(level, f"Output type: {type(output_data)}")

def log_validation_errors(errors: list[str], level: str = "ERROR") -> None:
    """Log validation errors.
    
    Args:
        errors: List of error messages
        level: Logging level
    """
    if not errors:
        return
    
    logger.log(level, "Validation errors:")
    for error in errors:
        logger.log(level, f"  - {error}")

class StageLogger:
    """Context manager for logging stage execution.
    
    Example:
        with StageLogger("process_image") as log:
            result = process_image(data)
            log.set_result(result)
    """
    
    def __init__(
        self,
        stage_name: str,
        level: str = "DEBUG",
        log_result: bool = True
    ):
        self.stage_name = stage_name
        self.level = level
        self.log_result = log_result
        self.result: Any = None
    
    def __enter__(self) -> 'StageLogger':
        logger.log(self.level, f"Starting stage: {self.stage_name}")
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if exc_type is not None:
            logger.error(f"Error in stage {self.stage_name}: {exc_val}")
        else:
            logger.log(self.level, f"Completed stage: {self.stage_name}")
            if self.log_result and self.result is not None:
                logger.log(self.level, f"Result type: {type(self.result)}")
    
    def set_result(self, result: Any) -> None:
        """Set the result of the stage execution."""
        self.result = result
        
    def debug(self, message: str) -> None:
        """Log a debug message."""
        logger.debug(f"[{self.stage_name}] {message}")
        
    def info(self, message: str) -> None:
        """Log an info message."""
        logger.info(f"[{self.stage_name}] {message}")
        
    def warning(self, message: str) -> None:
        """Log a warning message."""
        logger.warning(f"[{self.stage_name}] {message}")
        
    def error(self, message: str) -> None:
        """Log an error message."""
        logger.error(f"[{self.stage_name}] {message}")
        
    def critical(self, message: str) -> None:
        """Log a critical message."""
        logger.critical(f"[{self.stage_name}] {message}") 