"""Core pipeline functionality for composing and executing data transformations with strong validation."""

from typing import TypeVar, Generic, Callable, Any, Optional, Dict, List, Union, get_type_hints
from dataclasses import dataclass
from functools import wraps
import inspect
from .validation import validate_type
from .timing import timing_block
from .logging import log_dict, log_stage_execution, StageLogger
from loguru import logger

Input = TypeVar('Input')
Output = TypeVar('Output')
NextOutput = TypeVar('NextOutput')

class PipelineError(Exception):
    """Base exception for pipeline errors."""
    def __init__(self, message: str, details: Optional[Dict[str, Any]] = None, cause: Optional[Exception] = None):
        """Initialize pipeline error.
        
        Args:
            message: Error message
            details: Optional error details
            cause: Optional causing exception
        """
        self.details = details or {}
        self.cause = cause
        super().__init__(f"{message}\nDetails: {self.details}")

@dataclass
class StepMetadata:
    """Metadata for a pipeline step."""
    name: str
    description: Optional[str] = None
    input_type: Optional[type] = None
    output_type: Optional[type] = None
    is_pure: bool = True

    def validate(self) -> None:
        """Validate metadata consistency."""
        if not self.name:
            raise PipelineError("Step name cannot be empty")

class Step:
    """A step in a pipeline with type validation and metadata."""
    
    def __init__(
        self, 
        func: Callable,
        metadata: Optional[StepMetadata] = None,
        is_pure: bool = True
    ):
        """Initialize a step with a function and optional metadata.
        
        If metadata is not provided, it will be automatically extracted from the function.
        
        Args:
            func: The function to execute in this step
            metadata: Optional metadata for the step. If not provided, will be extracted
            is_pure: Whether the function is pure (no side effects)
        """
        self.func = func
        
        # If metadata not provided, extract it automatically
        if metadata is None:
            metadata = self._extract_metadata(func, is_pure)
        
        self.metadata = metadata
        self.metadata.validate()
    
    def _extract_metadata(self, func: Callable, is_pure: bool) -> StepMetadata:
        """Extract metadata from a function's signature and docstring.
        
        Args:
            func: The function to analyze
            is_pure: Whether the function is pure
            
        Returns:
            Extracted StepMetadata
            
        Raises:
            PipelineError: If type hints are missing or invalid
        """
        # Get function signature
        sig = inspect.signature(func)
        
        # Get type hints
        hints = get_type_hints(func)
        
        # Validate parameter has type hint
        params = list(sig.parameters.values())
        if len(params) != 1:
            raise PipelineError(
                f"Function {func.__name__} must take exactly one parameter",
                details={"function": func.__name__, "params": len(params)}
            )
        
        # Get input and output types
        param_name = next(iter(sig.parameters))
        input_type = hints.get(param_name)
        output_type = hints.get('return')
        
        # Check for missing type hints
        if not input_type and not output_type:
            raise PipelineError(
                f"Function {func.__name__} is missing type hints",
                details={"function": func.__name__}
            )
        
        # Check for missing parameter type hint
        if not input_type:
            raise PipelineError(
                f"Function {func.__name__} parameter is missing input type hint",
                details={
                    "function": func.__name__,
                    "parameter": param_name
                }
            )
        
        # Check for missing return type hint
        if not output_type:
            raise PipelineError(
                f"Function {func.__name__} is missing return type hint",
                details={
                    "function": func.__name__
                }
            )
        
        # Extract description from docstring
        doc = inspect.getdoc(func)
        description = doc.split('\n')[0] if doc else f"Step using {func.__name__}"
        
        # Create metadata
        return StepMetadata(
            name=func.__name__,
            description=description,
            input_type=input_type,
            output_type=output_type,
            is_pure=is_pure
        )
    
    def __call__(self, value: Any) -> Any:
        """Execute the step with validation.
        
        Args:
            value: Input value to process
            
        Returns:
            Processed value
            
        Raises:
            PipelineError: If validation fails or step execution fails
        """
        try:
            # Validate input type
            if not validate_type(value, self.metadata.input_type):
                raise PipelineError(
                    "Invalid input type",
                    details={
                        "step": self.metadata.name,
                        "expected": str(self.metadata.input_type),
                        "got": str(type(value))
                    }
                )
            
            # Execute step
            result = self.func(value)
            
            # Validate output type
            if not validate_type(result, self.metadata.output_type):
                raise PipelineError(
                    "Invalid output type",
                    details={
                        "step": self.metadata.name,
                        "expected": str(self.metadata.output_type),
                        "got": str(type(result))
                    }
                )
            
            return result
            
        except PipelineError:
            raise
        except Exception as e:
            raise PipelineError(
                f"Error in step {self.metadata.name}",
                details={"error": str(e)},
                cause=e
            )

class Pipeline(Generic[Input, Output]):
    """A pipeline for processing data through multiple steps with strong validation."""
    
    def __init__(self):
        self.steps: List[Step] = []
    
    def validate(self) -> None:
        """Validate the entire pipeline configuration."""
        if not self.steps:
            return
        
        # Validate step connections
        for i in range(len(self.steps) - 1):
            current = self.steps[i]
            next_step = self.steps[i + 1]
            
            # Check if output type of current step is compatible with input type of next step
            if current.metadata.output_type and next_step.metadata.input_type:
                if not validate_type(None, next_step.metadata.input_type) and \
                   not validate_type(None, current.metadata.output_type):
                    # Both types are concrete (not Any), check compatibility
                    if current.metadata.output_type != next_step.metadata.input_type:
                        raise PipelineError(
                            "Pipeline type mismatch",
                            details={
                                "step": f"{current.metadata.name} -> {next_step.metadata.name}",
                                "output_type": str(current.metadata.output_type),
                                "input_type": str(next_step.metadata.input_type)
                            }
                        )
    
    def then(self, step_or_func: Union[Step, Callable]) -> 'Pipeline':
        """Add a step to the pipeline."""
        if isinstance(step_or_func, Step):
            step = step_or_func
        else:
            step = Step(step_or_func)
            
        if self.steps:
            current = self.steps[-1]
            if current.metadata.output_type and step.metadata.input_type:
                if current.metadata.output_type != step.metadata.input_type:
                    raise PipelineError(
                        "Pipeline type mismatch",
                        details={
                            "current_step": current.metadata.name,
                            "next_step": step.metadata.name,
                            "output_type": str(current.metadata.output_type),
                            "input_type": str(step.metadata.input_type)
                        }
                    )
                    
        self.steps.append(step)
        return self
    
    def process(self, data: Input) -> Output:
        """Process data through the pipeline."""
        if not self.steps:
            raise PipelineError("Pipeline has no steps")
            
        try:
            current = data
            for step in self.steps:
                with timing_block(f"Step: {step.metadata.name}"):
                    try:
                        current = step(current)
                    except PipelineError as e:
                        # Re-raise PipelineError with original message and details
                        if "Invalid input type" in str(e) or "Custom error" in str(e):
                            # Preserve validation and custom error messages
                            raise e
                        # Wrap other PipelineErrors with "Pipeline execution failed"
                        raise PipelineError(
                            "Pipeline execution failed",
                            details=e.details,
                            cause=e
                        )
                    except Exception as e:
                        # Wrap other exceptions with "Pipeline execution failed"
                        raise PipelineError(
                            "Pipeline execution failed",
                            details={
                                "step": step.metadata.name,
                                "error": str(e)
                            },
                            cause=e
                        )
            return current
        except Exception as e:
            if isinstance(e, PipelineError):
                if "Invalid input type" in str(e) or "Custom error" in str(e):
                    # Preserve validation and custom error messages
                    raise e
                # Wrap other PipelineErrors with "Pipeline execution failed"
                raise PipelineError(
                    "Pipeline execution failed",
                    details=e.details,
                    cause=e
                )
            raise PipelineError(
                "Pipeline execution failed",
                details={
                    "step": step.metadata.name,
                    "error": str(e)
                },
                cause=e
            )
    
    def __str__(self) -> str:
        """String representation of the pipeline."""
        if not self.steps:
            return "Pipeline(empty)"
        return "Pipeline(" + " -> ".join(s.metadata.name for s in self.steps) + ")"

    def __call__(self, data: Input) -> Output:
        """Process data through the pipeline."""
        try:
            return self.process(data)
        except Exception as e:
            if isinstance(e, PipelineError):
                if "Invalid input type" in str(e) or "Custom error" in str(e):
                    # Preserve validation and custom error messages
                    raise e
                # Wrap other PipelineErrors with "Pipeline execution failed"
                raise PipelineError(
                    "Pipeline execution failed",
                    details=e.details,
                    cause=e
                )
            raise PipelineError(
                "Pipeline execution failed",
                details={
                    "error": str(e)
                },
                cause=e
            ) 