"""Utilities for timing and performance monitoring."""

import time
from contextlib import contextmanager
from typing import Optional, Generator
from loguru import logger

@contextmanager
def timing_block(name: str, level: str = "DEBUG") -> Generator[None, None, None]:
    """Context manager for timing code blocks.
    
    Args:
        name: Name of the operation being timed
        level: Logging level to use (DEBUG, INFO, etc.)
        
    Example:
        with timing_block("Loading model"):
            model.load_weights("weights.h5")
    """
    start = time.perf_counter()
    try:
        yield
    finally:
        duration = time.perf_counter() - start
        logger.log(level, f"{name} took {duration:.3f} seconds")

class Timer:
    """Utility class for timing operations.
    
    Attributes:
        total_time: Total accumulated time
        last_duration: Duration of the last timed operation
        is_running: Whether the timer is currently running
    """
    
    def __init__(self):
        self.total_time: float = 0.0
        self.last_duration: Optional[float] = None
        self.is_running: bool = False
        self._start: Optional[float] = None
    
    def start(self) -> None:
        """Start the timer."""
        if self.is_running:
            logger.warning("Timer is already running")
            return
        self._start = time.perf_counter()
        self.is_running = True
    
    def stop(self) -> float:
        """Stop the timer and return the duration."""
        if not self.is_running:
            logger.warning("Timer is not running")
            return 0.0
        
        if self._start is None:
            logger.error("Timer start time is None")
            return 0.0
        
        duration = time.perf_counter() - self._start
        self.total_time += duration
        self.last_duration = duration
        self.is_running = False
        self._start = None
        return duration
    
    def reset(self) -> None:
        """Reset the timer."""
        self.total_time = 0.0
        self.last_duration = None
        self.is_running = False
        self._start = None
    
    @contextmanager
    def time_operation(self, name: str, level: str = "DEBUG") -> Generator[None, None, None]:
        """Context manager for timing an operation.
        
        Args:
            name: Name of the operation being timed
            level: Logging level to use
            
        Example:
            timer = Timer()
            with timer.time_operation("Processing"):
                process_data()
        """
        self.start()
        try:
            yield
        finally:
            duration = self.stop()
            logger.log(level, f"{name} took {duration:.3f} seconds") 