"""Utilities for type and requirement validation."""

from typing import Any, Dict, Type, Optional, Union, get_args, get_origin
import inspect

def validate_type(value: Any, expected_type: Type) -> bool:
    """Validate that a value matches an expected type.
    
    Handles complex types like Union, Optional, etc.
    
    Args:
        value: The value to check
        expected_type: The expected type
        
    Returns:
        True if the value matches the expected type
    """
    if expected_type is Any:
        return True
    
    # Handle None/Optional
    if value is None:
        try:
            return type(None) in get_args(expected_type)
        except TypeError:
            # For simple types that don't support get_args
            return False
    
    # Get the origin type (e.g., list from List[int])
    origin = get_origin(expected_type)
    if origin is None:
        # Simple type
        return isinstance(value, expected_type)
    
    # Handle Union types
    if origin is Union:
        return any(validate_type(value, arg) for arg in get_args(expected_type))
    
    # Handle container types (List, Dict, etc.)
    if origin in (list, set):
        if not isinstance(value, origin):
            return False
        # Get the type parameter (e.g., int from List[int])
        type_args = get_args(expected_type)
        if not type_args:
            return True
        return all(validate_type(item, type_args[0]) for item in value)
    
    # Special handling for tuples
    if origin is tuple:
        if not isinstance(value, tuple):
            return False
        type_args = get_args(expected_type)
        # Handle empty tuple
        if not type_args:
            return True
        # Handle variable-length tuple (Tuple[int, ...])
        if len(type_args) == 2 and type_args[1] is Ellipsis:
            return all(validate_type(item, type_args[0]) for item in value)
        # Handle fixed-length tuple
        if len(value) != len(type_args):
            return False
        return all(validate_type(item, type_arg) 
                  for item, type_arg in zip(value, type_args))
    
    if origin is dict:
        if not isinstance(value, dict):
            return False
        type_args = get_args(expected_type)
        if not type_args:
            return True
        key_type, value_type = type_args
        return all(
            validate_type(k, key_type) and validate_type(v, value_type)
            for k, v in value.items()
        )
    
    # Other generic types
    return isinstance(value, origin)

def validate_requirements(
    requirements: Dict[str, Any],
    provided: Dict[str, Any]
) -> tuple[bool, list[str]]:
    """Validate that all requirements are met.
    
    Args:
        requirements: Dictionary of required values
        provided: Dictionary of provided values
        
    Returns:
        Tuple of (is_valid, list of error messages)
    """
    errors = []
    
    for key, required_value in requirements.items():
        if key not in provided:
            errors.append(f"Missing required value: {key}")
            continue
        
        provided_value = provided[key]
        
        # Type validation
        if isinstance(required_value, type):
            if not validate_type(provided_value, required_value):
                errors.append(
                    f"Type mismatch for {key}: expected {required_value}, "
                    f"got {type(provided_value)}"
                )
        
        # Value validation
        elif provided_value != required_value:
            errors.append(
                f"Value mismatch for {key}: expected {required_value}, "
                f"got {provided_value}"
            )
    
    return len(errors) == 0, errors

def validate_callable(
    func: Any,
    expected_input: Optional[Type] = None,
    expected_output: Optional[Type] = None
) -> tuple[bool, list[str]]:
    """Validate a callable's signature.
    
    Args:
        func: The callable to validate
        expected_input: Expected input type
        expected_output: Expected output type
        
    Returns:
        Tuple of (is_valid, list of error messages)
    """
    errors = []
    
    if not callable(func):
        errors.append(f"{func} is not callable")
        return False, errors
    
    # Get function signature
    try:
        sig = inspect.signature(func)
    except ValueError as e:
        errors.append(f"Could not inspect {func}: {e}")
        return False, errors
    
    # Validate input parameter
    if expected_input is not None:
        params = list(sig.parameters.values())
        if not params:
            errors.append(f"{func} takes no parameters, expected one")
        elif len(params) > 1:
            errors.append(f"{func} takes multiple parameters, expected one")
        else:
            param = params[0]
            if param.annotation == inspect.Parameter.empty:
                errors.append(f"{func} parameter is missing type annotation")
            elif param.annotation != expected_input:
                errors.append(
                    f"Input type mismatch: {func} expects {param.annotation}, "
                    f"required {expected_input}"
                )
    
    # Validate return type
    if expected_output is not None:
        return_annotation = sig.return_annotation
        if return_annotation == inspect.Parameter.empty:
            errors.append(f"{func} is missing return type annotation")
        elif return_annotation != expected_output:
            errors.append(
                f"Output type mismatch: {func} returns {return_annotation}, "
                f"required {expected_output}"
            )
    
    return len(errors) == 0, errors 