"""Tests for functional programming utilities."""

import pytest
from typing import List, Optional, Any, Dict, Union
from oteilo.functional import (
    compose, pipe, with_logging, log_args,
    curry, maybe, memoize, trace, Pipeline
)
from loguru import logger

def test_compose():
    """Test function composition."""
    def add_one(x: int) -> int:
        return x + 1
    
    def multiply_by_two(x: int) -> int:
        return x * 2
    
    def to_string(x: int) -> str:
        return str(x)
    
    # Test basic composition
    f = compose(to_string, multiply_by_two, add_one)
    assert f(3) == "8"  # (3 + 1) * 2 = 8
    
    # Test single function
    f = compose(to_string)
    assert f(42) == "42"
    
    # Test empty composition
    f = compose()
    assert f(42) == 42

def test_pipe():
    """Test function pipeline."""
    def add_one(x: int) -> int:
        return x + 1
    
    def multiply_by_two(x: int) -> int:
        return x * 2
    
    def to_string(x: int) -> str:
        return str(x)
    
    # Test basic pipeline
    f = pipe(add_one, multiply_by_two, to_string)
    assert f(3) == "8"  # (3 + 1) * 2 = 8
    
    # Test single function
    f = pipe(to_string)
    assert f(42) == "42"
    
    # Test empty pipeline
    f = pipe()
    assert f(42) == 42

def test_with_logging():
    """Test logging level decorator."""
    @with_logging(level="DEBUG")
    def example_func(x: int) -> int:
        return x * 2
    
    # Store original level
    original_level = logger.level
    
    # Call function and verify level is restored
    result = example_func(21)
    assert result == 42
    assert logger.level == original_level

def test_log_args():
    """Test argument logging decorator."""
    @log_args(level="DEBUG")
    def add(x: int, y: int) -> int:
        return x + y
    
    # Call function and verify it works
    result = add(2, 3)
    assert result == 5
    
    # Call with kwargs
    result = add(x=2, y=3)
    assert result == 5

def test_curry():
    """Test function currying."""
    @curry
    def add(x: int, y: int, z: int) -> int:
        return x + y + z
    
    # Test full application
    assert add(1, 2, 3) == 6
    
    # Test partial application
    add_one = add(1)
    assert add_one(2, 3) == 6
    
    add_one_two = add_one(2)
    assert add_one_two(3) == 6
    
    # Test with keyword arguments
    assert add(x=1)(y=2)(z=3) == 6
    assert add(1, y=2)(3) == 6

def test_maybe():
    """Test maybe wrapper for handling None values."""
    @maybe
    def double(x: int) -> int:
        return x * 2
    
    # Test with valid input
    assert double(2) == 4
    
    # Test with None input
    assert double(None) is None
    
    # Test with invalid input type
    with pytest.raises(TypeError):
        double("not an int")
    
    # Test chaining maybe functions
    @maybe
    def add_one(x: int) -> int:
        return x + 1
    
    result = compose(add_one, double)(2)  # (2 * 2) + 1
    assert result == 5
    
    result = compose(add_one, double)(None)
    assert result is None

def test_memoize():
    """Test function memoization."""
    call_count = 0
    
    @memoize
    def fibonacci(n: int) -> int:
        nonlocal call_count
        call_count += 1
        if n < 2:
            return n
        return fibonacci(n-1) + fibonacci(n-2)
    
    # First call should compute
    result1 = fibonacci(5)
    initial_calls = call_count
    
    # Second call should use cache
    result2 = fibonacci(5)
    assert result1 == result2
    assert call_count == initial_calls  # Call count shouldn't increase
    
    # Test with different args
    fibonacci(6)  # Should increase call count
    assert call_count > initial_calls
    
    # Test with keyword arguments
    @memoize
    def greet(name: str, greeting: str = "Hello") -> str:
        nonlocal call_count
        call_count += 1
        return f"{greeting}, {name}!"
    
    # Test that different kwarg combinations are cached separately
    count_before = call_count
    assert greet("World") == "Hello, World!"
    assert greet("World", greeting="Hi") == "Hi, World!"
    assert call_count == count_before + 2
    
    # Repeat calls should use cache
    greet("World")
    greet("World", greeting="Hi")
    assert call_count == count_before + 2

def test_trace():
    """Test function tracing with different configurations."""
    # Test simple timing only (default)
    @trace
    def simple_func(x: int) -> int:
        return x * 2
    
    result = simple_func(21)
    assert result == 42
    
    # Test full tracing with DEBUG level
    @trace(level="DEBUG")
    def typed_func(x: int, y: str = "default") -> Dict[str, Any]:
        return {"result": x, "message": y}
    
    result = typed_func(42)
    assert result == {"result": 42, "message": "default"}
    
    # Test custom configuration
    @trace(level="INFO", timing=True, args=False)
    def custom_func(x: int) -> int:
        return x * 2
    
    result = custom_func(21)
    assert result == 42
    
    # Test class method tracing
    class DataProcessor:
        def __init__(self, multiplier: int):
            self.multiplier = multiplier
        
        @trace(level="DEBUG")
        def process(self, x: int) -> int:
            return x * self.multiplier
    
    processor = DataProcessor(2)
    result = processor.process(21)
    assert result == 42
    
    # Test error handling
    @trace(level="DEBUG")
    def failing_func(x: int) -> None:
        raise ValueError("Test error")
    
    with pytest.raises(ValueError):
        failing_func(42)

def test_trace_type_handling():
    """Test trace decorator type handling."""
    @trace(level="DEBUG", types=True)
    def func_with_types(x: int, y: Optional[str] = None) -> Dict[str, Any]:
        return {"x": x, "y": y or "default"}
    
    result = func_with_types(42)
    assert result == {"x": 42, "y": "default"}
    
    # Test with Union types
    @trace(level="DEBUG", types=True)
    def func_with_union(x: Union[int, str]) -> str:
        return str(x)
    
    result = func_with_union(42)
    assert result == "42"
    result = func_with_union("42")
    assert result == "42"

def test_trace_nested_functions():
    """Test trace decorator with nested function calls."""
    @trace(level="DEBUG")
    def outer(x: int) -> int:
        return inner(x * 2)
    
    @trace(level="DEBUG")
    def inner(x: int) -> int:
        return x * 2
    
    result = outer(10)
    assert result == 40

def test_trace_without_logging():
    """Test trace decorator with logging disabled."""
    @trace(level=None)
    def simple_func(x: int) -> int:
        return x * 2
    
    result = simple_func(21)
    assert result == 42
    
    # Test with timing disabled
    @trace(level=None, timing=False)
    def no_trace_func(x: int) -> int:
        return x * 2
    
    result = no_trace_func(21)
    assert result == 42

def test_with_logging_error_handling():
    """Test with_logging decorator error handling."""
    @with_logging(level="DEBUG")
    def failing_func():
        raise ValueError("Test error")
    
    # Verify that logger level is restored even if function fails
    original_level = logger.level
    with pytest.raises(ValueError):
        failing_func()
    assert logger.level == original_level

def test_log_args_error_handling():
    """Test log_args decorator error handling."""
    @log_args(level="DEBUG")
    def failing_func(x: int):
        raise ValueError("Test error")
    
    with pytest.raises(ValueError):
        failing_func(42)

def test_compose_error_handling():
    """Test compose function error handling."""
    def fail_on_two(x: int) -> int:
        if x == 2:
            raise ValueError("No twos allowed")
        return x
    
    def double(x: int) -> int:
        return x * 2
    
    f = compose(double, fail_on_two)
    assert f(1) == 2  # Works for 1
    with pytest.raises(ValueError):
        f(2)  # Fails for 2

def test_pipe_error_handling():
    """Test pipe function error handling."""
    def fail_on_two(x: int) -> int:
        if x == 2:
            raise ValueError("No twos allowed")
        return x
    
    def double(x: int) -> int:
        return x * 2
    
    f = pipe(fail_on_two, double)
    assert f(1) == 2  # Works for 1
    with pytest.raises(ValueError):
        f(2)  # Fails for 2

def test_curry_complex_cases():
    """Test curry decorator with complex cases."""
    @curry
    def format_person(name: str, age: int, city: str = "Unknown") -> Dict[str, Union[str, int]]:
        return {"name": name, "age": age, "city": city}
    
    # Test with default argument
    assert format_person("Alice")(25) == {"name": "Alice", "age": 25, "city": "Unknown"}
    
    # Test with keyword arguments in different orders
    assert format_person("Bob", city="London")(30) == {"name": "Bob", "age": 30, "city": "London"}
    assert format_person(age=35)("Charlie", "Paris") == {"name": "Charlie", "age": 35, "city": "Paris"}
    
    # Test with all arguments at once
    assert format_person("David", 40, "Berlin") == {"name": "David", "age": 40, "city": "Berlin"}
    
    # Test with mixed positional and keyword arguments
    assert format_person("Eve", city="Madrid")(45) == {"name": "Eve", "age": 45, "city": "Madrid"}

def test_maybe_complex_cases():
    """Test maybe wrapper with complex cases."""
    @maybe
    def process_dict(data: Dict[str, Any]) -> Optional[str]:
        return data.get("name", "").upper()
    
    # Test with valid input
    assert process_dict({"name": "alice"}) == "ALICE"
    
    # Test with None input
    assert process_dict(None) is None
    
    # Test with missing key
    assert process_dict({}) == ""
    
    # Test with invalid input type
    with pytest.raises(TypeError):
        process_dict("not a dict")
    
    # Test chaining maybe functions with complex types
    @maybe
    def extract_first_char(s: str) -> Optional[str]:
        return s[0] if s else None
    
    pipeline = compose(extract_first_char, process_dict)
    assert pipeline({"name": "bob"}) == "B"
    assert pipeline({"name": ""}) is None
    assert pipeline(None) is None

def test_memoize_complex_cases():
    """Test memoize decorator with complex cases."""
    call_count = 0
    
    @memoize
    def process_list(items: List[int], multiplier: int = 1) -> List[int]:
        nonlocal call_count
        call_count += 1
        return [x * multiplier for x in items]
    
    # Test with immutable arguments
    result1 = process_list([1, 2, 3], 2)
    result2 = process_list([1, 2, 3], 2)
    assert result1 == result2
    assert call_count == 1  # Should use cache
    
    # Test with different argument combinations
    process_list([1, 2, 3], 3)  # Different multiplier
    assert call_count == 2
    
    process_list([1, 2], 2)  # Different list
    assert call_count == 3
    
    # Test with keyword arguments
    process_list(items=[1, 2, 3], multiplier=2)  # Should use cache
    assert call_count == 3
    
    process_list(multiplier=2, items=[1, 2, 3])  # Should use cache
    assert call_count == 3

def test_trace_complex_cases():
    """Test trace decorator with complex cases."""
    @trace
    def process_data(data: Dict[str, Any], raise_error: bool = False) -> Dict[str, Any]:
        if raise_error:
            raise ValueError("Processing failed")
        return {k.upper(): v * 2 if isinstance(v, int) else v.upper()
                for k, v in data.items()}
    
    # Test successful execution
    result = process_data({"name": "alice", "age": 25})
    assert result == {"NAME": "ALICE", "AGE": 50}
    
    # Test with error
    with pytest.raises(ValueError):
        process_data({"name": "bob"}, raise_error=True)

class TestFunctionalPipeline:
    """Tests for the functional Pipeline class."""
    
    def test_empty_pipeline(self):
        """Test empty pipeline behavior."""
        pipeline = Pipeline()
        result = list(pipeline.process([1, 2, 3]))
        assert result == [1, 2, 3]
    
    def test_map_operation(self):
        """Test map operation."""
        pipeline = Pipeline().map(str)
        result = list(pipeline.process([1, 2, 3]))
        assert result == ["1", "2", "3"]
    
    def test_filter_operation(self):
        """Test filter operation."""
        pipeline = Pipeline().filter(lambda x: x % 2 == 0)
        result = list(pipeline.process([1, 2, 3, 4]))
        assert result == [2, 4]
    
    def test_reduce_operation(self):
        """Test reduce operation."""
        # Without initial value
        pipeline = Pipeline().reduce(lambda x, y: x + y)
        result = pipeline.process([1, 2, 3, 4])
        assert result == 10
        
        # With initial value
        pipeline = Pipeline().reduce(lambda x, y: x + y, initial=0)
        result = pipeline.process([1, 2, 3, 4])
        assert result == 10
    
    def test_validate_operation(self):
        """Test validate operation."""
        pipeline = Pipeline().validate(int)
        result = list(pipeline.process([1, "2", 3, None, 4.0]))
        assert result == [1, 3]
    
    def test_chained_operations(self):
        """Test chaining multiple operations."""
        pipeline = (Pipeline()
            .map(str)
            .filter(lambda x: len(x) == 1)
            .map(lambda x: int(x) * 2))
        
        result = list(pipeline.process([1, 22, 3, 444]))
        assert result == [2, 6]
    
    def test_pipeline_with_none_values(self):
        """Test pipeline handling of None values."""
        pipeline = (Pipeline()
            .map(maybe(lambda x: x * 2))
            .filter(lambda x: x is not None)
            .map(str))
        
        result = list(pipeline.process([1, None, 2, None, 3]))
        assert result == ["2", "4", "6"]
    
    def test_pipeline_error_handling(self):
        """Test pipeline error handling."""
        def failing_func(x: int) -> int:
            if x == 2:
                raise ValueError("Test error")
            return x * 2
        
        pipeline = Pipeline().map(failing_func)
        with pytest.raises(ValueError):
            list(pipeline.process([1, 2, 3]))
    
    def test_pipeline_with_custom_name_and_level(self):
        """Test pipeline with custom name and logging level."""
        pipeline = Pipeline(name="test_pipe", level="INFO")
        result = list(pipeline.map(str).process([1, 2, 3]))
        assert result == ["1", "2", "3"]
    
    def test_pipeline_reduce_error_handling(self):
        """Test reduce operation error handling."""
        # Test reduce with empty input
        pipeline = Pipeline().reduce(lambda x, y: x + y)
        with pytest.raises(TypeError):
            pipeline.process([])
        
        # Test reduce with single item
        result = pipeline.process([42])
        assert result == 42
        
        # Test reduce with initial value and empty input
        pipeline = Pipeline().reduce(lambda x, y: x + y, initial=0)
        assert pipeline.process([]) == 0
    
    def test_pipeline_validation_error_handling(self):
        """Test validation error handling in pipeline."""
        def validate_positive(x: int) -> bool:
            return x > 0
        
        pipeline = Pipeline().validate(int).filter(validate_positive)
        result = list(pipeline.process([1, "2", -3, None, 4.0, 5]))
        assert result == [1, 5]  # Only positive integers
    
    def test_pipeline_map_with_type_validation(self):
        """Test map operation with type validation."""
        @maybe
        def safe_int(x: Any) -> Optional[int]:
            try:
                return int(x)
            except (ValueError, TypeError):
                return None
        
        pipeline = (Pipeline()
            .map(safe_int)
            .filter(lambda x: x is not None)
            .map(lambda x: x * 2))
        
        result = list(pipeline.process(["1", "not a number", "2", None, "3"]))
        assert result == [2, 4, 6] 