"""Tests for I/O utilities."""

import pytest
from pathlib import Path
import json
from PIL import Image
import tempfile
from oteilo.io import (
    ensure_path,
    load_json,
    save_json,
    load_image,
    save_image,
    temp_path,
    get_file_info
)

@pytest.fixture
def temp_dir():
    """Create a temporary directory for testing."""
    with tempfile.TemporaryDirectory() as tmpdir:
        yield Path(tmpdir)

@pytest.fixture
def sample_json_file(temp_dir):
    """Create a sample JSON file."""
    data = {"name": "test", "value": 42}
    path = temp_dir / "test.json"
    with open(path, 'w', encoding='utf-8') as f:
        json.dump(data, f)
    return path

@pytest.fixture
def sample_image_file(temp_dir):
    """Create a sample image file."""
    image = Image.new('RGB', (100, 100), color='red')
    path = temp_dir / "test.png"
    image.save(path)
    return path

def test_ensure_path(temp_dir):
    """Test ensure_path function."""
    # Test with existing path
    path = ensure_path(temp_dir)
    assert isinstance(path, Path)
    assert path.exists()
    
    # Test with non-existent path
    with pytest.raises(FileNotFoundError):
        ensure_path(temp_dir / "nonexistent")
    
    # Test with string path
    path = ensure_path(str(temp_dir))
    assert isinstance(path, Path)
    assert path.exists()

def test_load_json(sample_json_file):
    """Test load_json function."""
    # Test loading valid JSON
    data = load_json(sample_json_file)
    assert data == {"name": "test", "value": 42}
    
    # Test with non-existent file
    with pytest.raises(FileNotFoundError):
        load_json(sample_json_file.parent / "nonexistent.json")
    
    # Test with invalid JSON
    invalid_json = sample_json_file.parent / "invalid.json"
    invalid_json.write_text("{invalid")
    with pytest.raises(json.JSONDecodeError):
        load_json(invalid_json)

def test_save_json(temp_dir):
    """Test save_json function."""
    data = {"name": "test", "value": 42}
    path = temp_dir / "output.json"
    
    # Test basic save
    save_json(data, path)
    assert path.exists()
    loaded = load_json(path)
    assert loaded == data
    
    # Test with different indentation
    save_json(data, path, indent=4)
    content = path.read_text()
    assert "    " in content  # Check indentation
    
    # Test overwriting
    new_data = {"updated": True}
    save_json(new_data, path)
    loaded = load_json(path)
    assert loaded == new_data

def test_load_image(sample_image_file):
    """Test load_image function."""
    # Test loading RGB image
    image = load_image(sample_image_file)
    assert isinstance(image, Image.Image)
    assert image.mode == 'RGB'
    assert image.size == (100, 100)
    
    # Test with non-RGB image
    grayscale = Image.new('L', (100, 100))
    gray_path = sample_image_file.parent / "gray.png"
    grayscale.save(gray_path)
    
    # Test with convert_to_rgb=True
    image = load_image(gray_path, convert_to_rgb=True)
    assert image.mode == 'RGB'
    
    # Test with convert_to_rgb=False
    image = load_image(gray_path, convert_to_rgb=False)
    assert image.mode == 'L'
    
    # Test with non-existent file
    with pytest.raises(FileNotFoundError):
        load_image(sample_image_file.parent / "nonexistent.png")

def test_save_image(temp_dir, sample_image_file):
    """Test save_image function."""
    original = load_image(sample_image_file)
    
    # Test basic save
    output_path = temp_dir / "output.png"
    save_image(original, output_path)
    assert output_path.exists()
    loaded = load_image(output_path)
    assert loaded.mode == original.mode
    assert loaded.size == original.size
    
    # Test with explicit format
    jpg_path = temp_dir / "output.jpg"
    save_image(original, jpg_path, format='JPEG')
    assert jpg_path.exists()
    loaded = load_image(jpg_path)
    assert loaded.mode == original.mode
    assert loaded.size == original.size

def test_temp_path():
    """Test temp_path context manager."""
    # Test basic usage
    with temp_path() as path:
        assert isinstance(path, Path)
        assert path.exists()
        # Write something to the file
        path.write_text("test")
        assert path.read_text() == "test"
    # File should be cleaned up
    assert not path.exists()
    
    # Test with suffix
    with temp_path(suffix='.txt') as path:
        assert path.suffix == '.txt'
        assert path.exists()
    assert not path.exists()
    
    # Test cleanup on error
    try:
        with temp_path() as path:
            assert path.exists()
            raise RuntimeError("Test error")
    except RuntimeError:
        assert not path.exists()  # Should still clean up

def test_get_file_info(temp_dir):
    """Test get_file_info function."""
    # Test with file
    file_path = temp_dir / "test.txt"
    file_path.write_text("test content")
    
    info = get_file_info(file_path)
    assert info['name'] == "test.txt"
    assert info['extension'] == ".txt"
    assert info['size'] > 0
    assert info['created'] > 0
    assert info['modified'] > 0
    assert info['is_file'] is True
    assert info['is_dir'] is False
    
    # Test with directory
    dir_info = get_file_info(temp_dir)
    assert dir_info['is_file'] is False
    assert dir_info['is_dir'] is True
    
    # Test with non-existent path
    with pytest.raises(FileNotFoundError):
        get_file_info(temp_dir / "nonexistent") 