"""Tests for the pipeline module."""

import pytest
from typing import List, Dict, Any, Optional, Union
from oteilo.pipeline import Pipeline, Step, StepMetadata, PipelineError

# Test fixtures
@pytest.fixture
def sample_data() -> Dict[str, Any]:
    """Sample data for testing."""
    return {"value": 42, "text": "hello"}

@pytest.fixture
def typed_step_functions():
    """Collection of typed step functions for testing."""
    
    def int_to_str(x: int) -> str:
        return str(x)
    
    def str_to_int(x: str) -> int:
        return int(x)
    
    def str_to_list(x: str) -> List[str]:
        return list(x)
    
    def untyped_func(x):
        return x
    
    def multi_param_func(x: int, y: int) -> int:
        return x + y
    
    return {
        "int_to_str": int_to_str,
        "str_to_int": str_to_int,
        "str_to_list": str_to_list,
        "untyped": untyped_func,
        "multi_param": multi_param_func
    }

# StepMetadata Tests
class TestStepMetadata:
    
    def test_valid_metadata(self):
        """Test creating valid metadata."""
        metadata = StepMetadata(
            name="test_step",
            description="Test step",
            input_type=int,
            output_type=str
        )
        metadata.validate()  # Should not raise
        
        assert metadata.name == "test_step"
        assert metadata.description == "Test step"
        assert metadata.input_type == int
        assert metadata.output_type == str
        assert metadata.is_pure is True
    
    def test_empty_name_raises(self):
        """Test that empty name raises error."""
        with pytest.raises(PipelineError) as exc_info:
            StepMetadata(name="").validate()
        
        assert "Step name cannot be empty" in str(exc_info.value)

# Step Tests
class TestStep:
    
    def test_valid_step_creation(self, typed_step_functions):
        """Test creating a valid step."""
        step = Step(typed_step_functions["int_to_str"])
        
        assert step.metadata.name == "int_to_str"
        assert step.metadata.input_type == int
        assert step.metadata.output_type == str
    
    def test_untyped_function_raises(self, typed_step_functions):
        """Test that untyped function raises error."""
        with pytest.raises(PipelineError) as exc_info:
            Step(typed_step_functions["untyped"])
        
        assert "missing type hints" in str(exc_info.value)
    
    def test_multi_param_function_raises(self, typed_step_functions):
        """Test that function with multiple parameters raises error."""
        with pytest.raises(PipelineError) as exc_info:
            Step(typed_step_functions["multi_param"])
        
        assert "must take exactly one parameter" in str(exc_info.value)
    
    def test_step_execution(self, typed_step_functions):
        """Test step execution with validation."""
        step = Step(typed_step_functions["int_to_str"])
        result = step(42)
        
        assert result == "42"
        
        # Test with invalid input type
        with pytest.raises(PipelineError) as exc_info:
            step("not an int")
        assert "Invalid input type" in str(exc_info.value)
    
    def test_step_execution_error_handling(self):
        """Test step execution error handling."""
        def failing_step(x: int) -> str:
            raise ValueError("Something went wrong")
        
        step = Step(failing_step)
        with pytest.raises(PipelineError) as exc_info:
            step(42)
        
        assert "Error in step" in str(exc_info.value)
        assert "Something went wrong" in str(exc_info.value)
    
    def test_step_with_custom_metadata(self):
        """Test step creation with custom metadata."""
        def func(x: int) -> str:
            """Convert int to string."""
            return str(x)
            
        metadata = StepMetadata(
            name="custom_step",
            description="Custom description",
            input_type=int,
            output_type=str,
            is_pure=False
        )
        
        step = Step(func, metadata=metadata)
        assert step.metadata.name == "custom_step"
        assert step.metadata.description == "Custom description"
        assert step.metadata.input_type == int
        assert step.metadata.output_type == str
        assert step.metadata.is_pure is False
    
    def test_step_docstring_handling(self):
        """Test that step uses function docstring as description."""
        def func(x: int) -> str:
            """This is a test docstring."""
            return str(x)
            
        step = Step(func)
        assert step.metadata.description == "This is a test docstring."
    
    def test_step_output_type_validation(self):
        """Test validation of step output type."""
        def incorrect_return(x: int) -> str:
            return 42  # type: ignore
        
        step = Step(incorrect_return)
        with pytest.raises(PipelineError) as exc_info:
            step(42)
        assert "Invalid output type" in str(exc_info.value)
        assert str(int) in str(exc_info.value)
        assert str(str) in str(exc_info.value)
    
    def test_step_type_extraction_edge_cases(self):
        """Test edge cases in type hint extraction."""
        def no_return_annotation(x: int):  # type: ignore
            return str(x)
        
        with pytest.raises(PipelineError) as exc_info:
            Step(no_return_annotation)
        assert "missing return type hint" in str(exc_info.value)
        
        def no_param_annotation(x) -> str:  # type: ignore
            return str(x)
        
        with pytest.raises(PipelineError) as exc_info:
            Step(no_param_annotation)
        assert "missing input type hint" in str(exc_info.value)
        
        def no_type_hints(x):  # type: ignore
            return str(x)
        
        with pytest.raises(PipelineError) as exc_info:
            Step(no_type_hints)
        assert "missing type hints" in str(exc_info.value)

# Pipeline Tests
class TestPipeline:
    
    def test_empty_pipeline(self):
        """Test empty pipeline behavior."""
        pipeline = Pipeline()
        assert str(pipeline) == "Pipeline(empty)"
        
        with pytest.raises(PipelineError) as exc_info:
            pipeline.process(42)
        assert "Pipeline has no steps" in str(exc_info.value)
    
    def test_single_step_pipeline(self, typed_step_functions):
        """Test pipeline with a single step."""
        pipeline = Pipeline().then(typed_step_functions["int_to_str"])
        result = pipeline.process(42)
        
        assert result == "42"
        assert str(pipeline) == "Pipeline(int_to_str)"
    
    def test_multi_step_pipeline(self, typed_step_functions):
        """Test pipeline with multiple compatible steps."""
        pipeline = (Pipeline()
            .then(typed_step_functions["int_to_str"])
            .then(typed_step_functions["str_to_list"]))
        
        result = pipeline.process(42)
        assert result == ['4', '2']
        assert str(pipeline) == "Pipeline(int_to_str -> str_to_list)"
    
    def test_incompatible_steps_raises(self, typed_step_functions):
        """Test that incompatible steps raise error."""
        pipeline = Pipeline().then(typed_step_functions["int_to_str"])
        
        with pytest.raises(PipelineError) as exc_info:
            pipeline.then(typed_step_functions["int_to_str"])
        
        assert "Pipeline type mismatch" in str(exc_info.value)
    
    def test_pipeline_execution_error_handling(self, typed_step_functions):
        """Test pipeline execution error handling."""
        def failing_step(x: str) -> int:
            raise ValueError("Step failed")
        
        pipeline = (Pipeline()
            .then(typed_step_functions["int_to_str"])
            .then(failing_step))
        
        with pytest.raises(PipelineError) as exc_info:
            pipeline.process(42)
        
        assert "Pipeline execution failed" in str(exc_info.value)
        assert "Step failed" in str(exc_info.value)
    
    def test_pipeline_with_preconstructed_step(self, typed_step_functions):
        """Test pipeline with a pre-constructed step."""
        step = Step(typed_step_functions["int_to_str"])
        pipeline = Pipeline().then(step)
        
        result = pipeline.process(42)
        assert result == "42"
    
    def test_pipeline_with_union_types(self):
        """Test pipeline with union types."""
        def step1(x: int) -> Union[int, str]:
            return str(x)
            
        def step2(x: Union[int, str]) -> str:
            return str(x)
            
        pipeline = Pipeline().then(step1).then(step2)
        result = pipeline.process(42)
        assert result == "42"
    
    def test_pipeline_error_reraising(self):
        """Test that PipelineError is re-raised without wrapping."""
        def step1(x: int) -> str:
            raise PipelineError("Custom error", {"custom": "details"})
            
        pipeline = Pipeline().then(step1)
        with pytest.raises(PipelineError) as exc_info:
            pipeline.process(42)
            
        assert "Custom error" in str(exc_info.value)
        assert "custom" in exc_info.value.details
        assert exc_info.value.details["custom"] == "details"
    
    def test_pipeline_validation_with_multiple_steps(self, typed_step_functions):
        """Test validation with multiple steps."""
        pipeline = Pipeline()
        
        # First step
        pipeline.then(typed_step_functions["int_to_str"])
        pipeline.validate()  # Should not raise
        
        # Second compatible step
        pipeline.then(typed_step_functions["str_to_list"])
        pipeline.validate()  # Should not raise
        
        # Try to add incompatible step
        with pytest.raises(PipelineError) as exc_info:
            pipeline.then(typed_step_functions["int_to_str"])
            pipeline.validate()
        assert "Pipeline type mismatch" in str(exc_info.value)
    
    def test_pipeline_with_any_type(self):
        """Test pipeline with Any type hints."""
        def any_to_str(x: Any) -> str:
            return str(x)
            
        def str_to_any(x: str) -> Any:
            return int(x)
            
        pipeline = Pipeline().then(any_to_str).then(str_to_any)
        result = pipeline.process(42)
        assert result == 42
    
    def test_pipeline_with_optional_types(self):
        """Test pipeline with Optional type hints."""
        def process_optional(x: Optional[int]) -> Optional[str]:
            if x is None:
                return None
            return str(x)
            
        def str_to_int(x: Optional[str]) -> int:
            if x is None:
                return 0
            return int(x)
            
        pipeline = Pipeline().then(process_optional).then(str_to_int)
        
        assert pipeline.process(42) == 42
        assert pipeline.process(None) == 0
    
    def test_pipeline_error_details(self):
        """Test error details in pipeline errors."""
        def failing_step(x: int) -> str:
            raise ValueError("Custom error message")
            
        pipeline = Pipeline().then(failing_step)
        
        with pytest.raises(PipelineError) as exc_info:
            pipeline.process(42)
            
        error = exc_info.value
        assert "Error in step" in str(error)
        assert "Custom error message" in str(error.details.get("error", ""))
    
    def test_pipeline_step_chaining_edge_cases(self):
        """Test edge cases in step chaining."""
        def identity(x: Any) -> Any:
            return x
            
        # Test chaining multiple identity functions
        pipeline = Pipeline()
        for _ in range(3):
            pipeline.then(identity)
            
        result = pipeline.process(42)
        assert result == 42
        assert str(pipeline) == "Pipeline(identity -> identity -> identity)"
    
    def test_pipeline_with_complex_types(self):
        """Test pipeline with complex type hints."""
        def list_to_dict(x: List[int]) -> Dict[str, int]:
            return {str(i): v for i, v in enumerate(x)}
            
        def dict_to_list(x: Dict[str, int]) -> List[int]:
            return list(x.values())
            
        pipeline = Pipeline().then(list_to_dict).then(dict_to_list)
        
        input_data = [1, 2, 3]
        result = pipeline.process(input_data)
        assert result == input_data

# Integration Tests
@pytest.mark.integration
class TestPipelineIntegration:
    
    def test_complex_data_transformation(self):
        """Test pipeline with complex data transformations."""
        def extract_value(data: Dict[str, Any]) -> int:
            return data["value"]
        
        def multiply_by_two(x: int) -> int:
            return x * 2
        
        def format_result(x: int) -> Dict[str, Any]:
            return {"result": x, "is_even": x % 2 == 0}
        
        pipeline = (Pipeline()
            .then(extract_value)
            .then(multiply_by_two)
            .then(format_result))
        
        input_data = {"value": 21, "extra": "ignored"}
        result = pipeline.process(input_data)
        
        assert result == {"result": 42, "is_even": True}
    
    def test_pipeline_with_validation(self):
        """Test pipeline with various validation scenarios."""
        def validate_positive(x: int) -> int:
            if x <= 0:
                raise ValueError("Value must be positive")
            return x
        
        def square(x: int) -> int:
            return x * x
        
        pipeline = (Pipeline()
            .then(validate_positive)
            .then(square))
        
        # Test valid input
        assert pipeline.process(5) == 25
        
        # Test invalid input
        with pytest.raises(PipelineError) as exc_info:
            pipeline.process(-1)
        assert "Value must be positive" in str(exc_info.value)
        
        # Test type validation
        with pytest.raises(PipelineError) as exc_info:
            pipeline.process("not an int")
        assert "Invalid input type" in str(exc_info.value) 