"""Tests for timing utilities."""

import pytest
import time
from oteilo.timing import timing_block, Timer

def test_timing_block():
    """Test timing_block context manager."""
    # Test basic usage
    with timing_block("test_operation"):
        time.sleep(0.1)  # Simulate work
    
    # Test with different log level
    with timing_block("test_operation", level="INFO"):
        time.sleep(0.1)
    
    # Test with error
    with pytest.raises(ValueError):
        with timing_block("failing_operation"):
            raise ValueError("Test error")

class TestTimer:
    """Tests for Timer class."""
    
    def test_initialization(self):
        """Test timer initialization."""
        timer = Timer()
        assert timer.total_time == 0.0
        assert timer.last_duration is None
        assert not timer.is_running
        assert timer._start is None
    
    def test_start_stop(self):
        """Test start and stop operations."""
        timer = Timer()
        
        # Test basic start/stop
        timer.start()
        assert timer.is_running
        assert timer._start is not None
        time.sleep(0.1)
        duration = timer.stop()
        
        assert not timer.is_running
        assert timer._start is None
        assert duration > 0
        assert timer.last_duration == duration
        assert timer.total_time == duration
        
        # Test multiple operations
        timer.start()
        time.sleep(0.1)
        first_duration = timer.stop()
        initial_total = timer.total_time
        
        timer.start()
        time.sleep(0.1)
        second_duration = timer.stop()
        
        # Verify that total_time increases by each duration
        assert timer.total_time > initial_total
        assert timer.last_duration == second_duration
        assert timer.total_time > first_duration
        assert timer.total_time > second_duration
    
    def test_reset(self):
        """Test reset operation."""
        timer = Timer()
        
        # Accumulate some time
        timer.start()
        time.sleep(0.1)
        timer.stop()
        
        assert timer.total_time > 0
        assert timer.last_duration is not None
        
        # Reset and verify
        timer.reset()
        assert timer.total_time == 0.0
        assert timer.last_duration is None
        assert not timer.is_running
        assert timer._start is None
    
    def test_time_operation(self):
        """Test time_operation context manager."""
        timer = Timer()
        
        # Test basic usage
        with timer.time_operation("test_operation"):
            time.sleep(0.1)
        
        assert not timer.is_running
        assert timer.last_duration is not None
        assert timer.total_time > 0
        
        # Test with different log level
        initial_total = timer.total_time
        with timer.time_operation("test_operation", level="INFO"):
            time.sleep(0.1)
        assert timer.total_time > initial_total
        
        # Test with error
        initial_total = timer.total_time
        try:
            with timer.time_operation("failing_operation"):
                raise ValueError("Test error")
        except ValueError:
            pass
        
        assert not timer.is_running
        assert timer.total_time > initial_total  # Time should still be recorded
    
    def test_error_handling(self):
        """Test error handling in Timer."""
        timer = Timer()
        
        # Test starting already running timer
        timer.start()
        timer.start()  # Should log warning
        timer.stop()
        
        # Test stopping non-running timer
        duration = timer.stop()  # Should log warning
        assert duration == 0.0
        
        # Test stopping with None start time
        timer.is_running = True
        timer._start = None
        duration = timer.stop()  # Should log error
        assert duration == 0.0
    
    def test_concurrent_operations(self):
        """Test handling of concurrent operations."""
        timer = Timer()
        
        # Start multiple operations without stopping
        timer.start()
        time.sleep(0.1)
        timer.start()  # Should log warning
        time.sleep(0.1)
        
        # Only one stop should be needed
        duration = timer.stop()
        assert duration > 0
        assert not timer.is_running 