"""Tests for validation utilities."""

import pytest
from typing import List, Dict, Set, Tuple, Union, Optional, Any
from oteilo.validation import validate_type, validate_requirements, validate_callable

# Test validate_type function
class TestValidateType:
    """Tests for validate_type function."""
    
    def test_simple_types(self):
        """Test validation of simple types."""
        assert validate_type(42, int)
        assert validate_type("hello", str)
        assert validate_type(3.14, float)
        assert not validate_type("42", int)
        assert not validate_type(42, str)
    
    def test_any_type(self):
        """Test validation against Any type."""
        assert validate_type(42, Any)
        assert validate_type("hello", Any)
        assert validate_type(None, Any)
        assert validate_type([1, 2, 3], Any)
    
    def test_optional_none(self):
        """Test validation of Optional types and None."""
        assert validate_type(None, Optional[int])
        assert validate_type(42, Optional[int])
        assert not validate_type(None, int)
        assert validate_type(None, Union[int, None])
    
    def test_union_types(self):
        """Test validation of Union types."""
        assert validate_type(42, Union[int, str])
        assert validate_type("hello", Union[int, str])
        assert not validate_type(3.14, Union[int, str])
        assert validate_type(42, Union[int, float, str])
    
    def test_list_validation(self):
        """Test validation of List types."""
        assert validate_type([1, 2, 3], List[int])
        assert validate_type([], List[int])
        assert not validate_type([1, "2", 3], List[int])
        assert validate_type(["a", "b"], List[str])
    
    def test_set_validation(self):
        """Test validation of Set types."""
        assert validate_type({1, 2, 3}, Set[int])
        assert validate_type(set(), Set[int])
        assert not validate_type({1, "2", 3}, Set[int])
        assert validate_type({"a", "b"}, Set[str])
    
    def test_tuple_validation(self):
        """Test validation of Tuple types."""
        # Fixed length tuples
        assert validate_type((1, "hello"), Tuple[int, str])
        assert not validate_type((1, 2), Tuple[int, str])
        assert not validate_type((1,), Tuple[int, str])
        
        # Variable length tuples
        assert validate_type((1, 2, 3), Tuple[int, ...])
        assert validate_type((), Tuple[int, ...])
        assert not validate_type((1, "2", 3), Tuple[int, ...])
    
    def test_dict_validation(self):
        """Test validation of Dict types."""
        assert validate_type({"a": 1, "b": 2}, Dict[str, int])
        assert validate_type({}, Dict[str, int])
        assert not validate_type({"a": "1"}, Dict[str, int])
        assert not validate_type({1: "a"}, Dict[str, int])

# Test validate_requirements function
class TestValidateRequirements:
    """Tests for validate_requirements function."""
    
    def test_basic_requirements(self):
        """Test basic requirement validation."""
        requirements = {"name": str, "age": int}
        provided = {"name": "John", "age": 30}
        is_valid, errors = validate_requirements(requirements, provided)
        assert is_valid
        assert not errors
    
    def test_missing_values(self):
        """Test handling of missing values."""
        requirements = {"name": str, "age": int}
        provided = {"name": "John"}
        is_valid, errors = validate_requirements(requirements, provided)
        assert not is_valid
        assert len(errors) == 1
        assert "Missing required value: age" in errors
    
    def test_type_mismatch(self):
        """Test type validation in requirements."""
        requirements = {"name": str, "age": int}
        provided = {"name": "John", "age": "30"}
        is_valid, errors = validate_requirements(requirements, provided)
        assert not is_valid
        assert len(errors) == 1
        assert f"Type mismatch for age: expected {int}, got {type('30')}" in errors
    
    def test_value_validation(self):
        """Test exact value validation."""
        requirements = {"status": "active", "version": 1}
        provided = {"status": "inactive", "version": 1}
        is_valid, errors = validate_requirements(requirements, provided)
        assert not is_valid
        assert len(errors) == 1
        assert f"Value mismatch for status: expected active, got inactive" in errors

# Test validate_callable function
class TestValidateCallable:
    """Tests for validate_callable function."""
    
    def test_valid_callable(self):
        """Test validation of valid callable."""
        def func(x: int) -> str:
            return str(x)
        
        is_valid, errors = validate_callable(func, int, str)
        print(f"\nErrors: {errors}")  # Debug output
        assert is_valid
        assert not errors
    
    def test_non_callable(self):
        """Test validation of non-callable object."""
        not_a_func = 42
        is_valid, errors = validate_callable(not_a_func)
        assert not is_valid
        assert len(errors) == 1
        assert "is not callable" in errors[0]
    
    def test_no_parameters(self):
        """Test validation of function with no parameters."""
        def func() -> str:
            return "hello"
        
        is_valid, errors = validate_callable(func, expected_input=int)
        assert not is_valid
        assert len(errors) == 1
        assert "takes no parameters" in errors[0]
    
    def test_multiple_parameters(self):
        """Test validation of function with multiple parameters."""
        def func(x: int, y: int) -> int:
            return x + y
        
        is_valid, errors = validate_callable(func, expected_input=int)
        assert not is_valid
        assert len(errors) == 1
        assert "takes multiple parameters" in errors[0]
    
    def test_type_mismatch(self):
        """Test validation of type mismatches."""
        def func(x: str) -> int:
            return len(x)
        
        is_valid, errors = validate_callable(func, expected_input=int, expected_output=str)
        assert not is_valid
        assert len(errors) == 2  # Both input and output type mismatches
        assert any("Input type mismatch" in error for error in errors)
        assert any("Output type mismatch" in error for error in errors) 